// ============================================================================
// File: questionGame.cpp
// Author: Jay Oblay (Frenchman extraordinaire)
// Last Modified:  Fall 2020
// Description: 
// This is a text based game where
// The player first thinks of an animal.
// The computer then asks a set of yes or no questions to
// discover what the animal is.
//
// If the computer guesses the correct animal all is good.
// If not the computer asks the user what animal he/she was thinking of
// and a yes or no question to determine the difference between
// what the computer guessed and what the animal actually was.
// These extra/new animals and related questions and answers 
// are stored in 2 support text files.
// See also the template file
//
// This serves an an illustrative (and fun) example of how to
// use the binary tree data structure.
// It also shows how to input and output to a file.
//
// Tasks: Find the TODOs, clean up this code
//
// Is this really using a tree class?  (it only declares the node type...)
// Could this code be restructured to use a tree class?
//
// TODO:
// This program stores out "new" questions and answers and animals
// to two text files
// Can you find a way to read them back in the next time the program is
// run? --- warning the current way things are output might not be
// good enough to work, do you need two files? or can it be done with 
// just one?
//
// TODO: clean up the comments, add comments where needed
//
// ============================================================================

// ---------------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <string>
#include "binaryTree.h"

using namespace std;

// ---------------------------------------------------------------------
// GLOBAL FUNCTION prototypes  
//                 C/C++ must declare the function before use
//                       definitions can come later
// ---------------------------------------------------------------------
void PrintInstructions();
btNode<string>* InitializeTree();
void learn(btNode<string>*& leafPtr);
void askAndMove(btNode<string>*& currentPtr);
void play(btNode<string>* currentPtr);
void eatLine();
bool inquire(const char query[]);

void saveTree(btNode<string>* rootPtr);
void saveNodeToFile(ofstream &treeFile, btNode<string>* curNode);

int readTree(btNode<string>*& rootPtr);
void makeTreeFromList(string *lineTree, int numNodes, 
                       btNode<string>*&curPtr, int &curIndex);

// ---------------------------------------------------------------------
// ----- main function, it all begins here
// ---------------------------------------------------------------------
int main()
{
   btNode<string> *taxonomyRootPtr;

   // Print out some instructions
   PrintInstructions();
   
   // Attempt to initialize the tree from a file
   // If unable to use the file, regress to a hardcoded setup
   if (readTree(taxonomyRootPtr) <= 0)
   {
      taxonomyRootPtr = InitializeTree();
   }
   
   // Use a do-while loop to guarantee at least play one time
   do
   {
      // TODO: call play function, send it taxonomyRootPtr
      play( XXXXX );
   }
   while (inquire("Shall we play again?"));
   
   // User has decided to quit, save the tree to a file 
   // TODO: call saveTree function, send it the root pointer
   XXXXXX ( XXXXX );
   
   cout << "Thank you for teaching me a thing or two." << endl;

   return 0;
}

// ---------------------------------------------------------------------
// PrintInstructions
// ---------------------------------------------------------------------
void PrintInstructions() 
{
   cout << "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n";
   cout << "Instructions: " << endl;
   cout << "   1. The player thinks of an animal." << endl;
   cout << "   2. Then the program tries to guess the animal." << endl;
   cout << "       -If the program cannot guess the animal it will ask you" << endl;
   cout << "        to enter the correct answer followed by a question to" << endl;
   cout << "        differentiate it from the last guessed animal" << endl;
   cout << "        and the correct animal." << endl;
   cout << "   3. The player will be asked to play again or exit." << endl << endl;
   cout << "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n";
}

// ---------------------------------------------------------------------
// InitializeTree
// the starting questions and answers
// ---------------------------------------------------------------------
btNode<string>* InitializeTree()
{
	btNode<string> *rootPtr;
	btNode<string> *childPtr;

    // Hard coded questions and answers
	const string rootQuestion("Are you a mammal?");
	const string leftQuestion("Are you bigger than a cat?");
	const string rightQuestion("Do you live underwater?");
	const string animal1("Kangaroo");
	const string animal2("Mouse");
	const string animal3("Shark");
	const string animal4("Seagull");

    // TODO: No coding for this item
    //       Draw out the decision tree being created below
    //       to see how it works (do not need to turn it in)
    //       Is "go left" a yes or a no answer?

    // Position rootPtr at the root question
	rootPtr = createNode(rootQuestion);
	
	childPtr = createNode(leftQuestion);      // make the left question
	childPtr->mp_left = createNode(animal1);  // make left answer
	childPtr->mp_right = createNode(animal2); // make right answer
	rootPtr->mp_left = childPtr;

	childPtr = createNode(rightQuestion);     // make the right question
	childPtr->mp_left = createNode(animal3);  // make left answer
	childPtr->mp_right = createNode(animal4); // make right answer
	rootPtr->mp_right = childPtr;
	
	return rootPtr;
}

// ---------------------------------------------------------------------
// readTree
// returns 0 if file is empty or unable to read from file
// ---------------------------------------------------------------------
int readTree(btNode<string>*& rootPtr)
{
   string lineTree[1000];
   ifstream treeFile ("WholeTree.txt");
   
   int numNodes = 0;
   
   if (treeFile.is_open())
   {
	   while ( getline (treeFile,lineTree[numNodes]) )
	   {
		   // cout << lineTree[numNodes] << endl;  // for debugging
		   numNodes++;
	   }
	   treeFile.close();
   }
  
   if (numNodes <= 0)  { cout << " No Whole Tree File, or empty\n"; return 0; }

   int curIndex = 0;
   makeTreeFromList(lineTree, numNodes, rootPtr, curIndex);
   
   return numNodes;
}

// ---------------------------------------------------------------------
// makeTreeFromList
// ---------------------------------------------------------------------
void makeTreeFromList(string *lineTree, int numNodes, 
                       btNode<string>*&curPtr, int &curIndex)
{
   if (lineTree == NULL)  { return; }
   if (numNodes <= 0 ) { return; }
   if (curIndex >= numNodes) { return; }
   if (lineTree[curIndex] == "null") 
   { 
      curPtr = NULL;
      curIndex++;
   }
   else
   {
      curPtr = createNode(lineTree[curIndex]);
      curIndex++;
      makeTreeFromList(lineTree, numNodes, curPtr->mp_left, curIndex);
      makeTreeFromList(lineTree, numNodes, curPtr->mp_right, curIndex);
   }
   
   return;
}

// ---------------------------------------------------------------------
// saveTree
// TODO: check the global prototype at the beginning of this file
//       see what paremeter(s) this function is supposed to take
// ---------------------------------------------------------------------
void saveTree( XXXXXXXX )
{
   if (rootPtr == NULL)   { return;  }
   	
   ofstream treeFile;
   treeFile.open("WholeTree.txt", ios::trunc);
   if(treeFile.is_open() == false)
   {
      cout << "Error opening WholeTree file!";
  }
   else   // if (treeFile.is_open())
   {
      // TODO: Call a function to: Save the Root Node to a file
      // May assume it recurses and saves all other nodes
      XXXX (treeFile, XXXXXX);
   }
   treeFile.close();

}

// ---------------------------------------------------------------------
// saveNodeToFile
// Recursively save a tree to a file ALREADY opened.
// First call to this function should send the root of the tree
// to be saved.
// ---------------------------------------------------------------------
void XXXXXXXX (ofstream &treeFile, btNode<string>* curNode)
{
	if (curNode == NULL) 
	{
		treeFile << "null" << endl; 
		return; 
	}
	
	// write current node's data
	treeFile << XXXXXXXX << endl;
	
	// TODO: recurse left tree then right tree 
	// hmmm... what type of traversal would this be?... 
	// pre-order... current node was already saved.
	saveNodeToFile(treeFile, XXXXXX );
	saveNodeToFile(XXXXXXXX, curNode->mp_right);
}

// ---------------------------------------------------------------------
// learn
// fills in users input and saves it
// inserts answer and question into the tree
// ---------------------------------------------------------------------
void learn(btNode<string>*& leafPtr)
{
	// TODO: no coding for this item
	//       Map out how the tree is being updated here
	//       Suggest go through it by hand
	//       Assume the 'hardcoded' tree from function: InitializeTree
	//       already exists and you are adding a new animal
	
	string guessAnimal;
	string correctAnimal;
	string newQuestion;

	guessAnimal = leafPtr->m_data;
	cout << "I give up. What are you? " << endl;
	getline(cin, correctAnimal); // enter the correct animal

	cout << "Please type a yes/no question that will distinguish a" << endl;
	cout << correctAnimal << " from a " << guessAnimal << "." << endl;
	getline(cin, newQuestion); // enter the correct question

	leafPtr->m_data = newQuestion;
	cout << "As a " << correctAnimal << ", " << newQuestion << endl;
	if (inquire("Please answer")) // if yes then set the correct animal to be the
	{                             // right side answer
		leafPtr->mp_left = createNode(correctAnimal);
		leafPtr->mp_right = createNode(guessAnimal);
	}
	else // other wise make it the left side answer
	{
		leafPtr->mp_left = createNode(guessAnimal);
		leafPtr->mp_right = createNode(correctAnimal);
	}

}

// ---------------------------------------------------------------------
// askAndMove
// moves pointer to next question 
// right if the question is answered with a no else left
// i.e. reposition the pointer
// ---------------------------------------------------------------------
void askAndMove(btNode<string>*& currentPtr)
{
	cout << currentPtr->m_data;
	if (inquire(" Please answer"))  // if answer is yes then position pointer left
		currentPtr = currentPtr->mp_left;
	else // otherwise move it to the right
		currentPtr = currentPtr->mp_right;
}

// ---------------------------------------------------------------------
// play
// Plays the game, see comments below
// ---------------------------------------------------------------------
void play(btNode<string>* currentPtr)
{
	cout << "Think of an animal, then press enter.";
	eatLine();//hit enter key
	
	while (!isLeaf(*currentPtr)) // if the current pointer is not empty
		askAndMove(currentPtr);  // then move the current pointer

	cout << ("My guess is " + currentPtr->m_data); // programs guess
	
	if (!inquire(". Am I right?"))   // if answers no
		learn(currentPtr);           // then put the answer and question into the tree
	else
		cout << "I knew it all along!" << endl; // otherwise tell user it knew it
}        

// ---------------------------------------------------------------------
// eatLine
// Clears a line of text from the input stream
// ---------------------------------------------------------------------
void eatLine()
{
	char next;
	
	do 

	cin.get(next);
	while (next != '\n');
}

// ---------------------------------------------------------------------
// inquire
// Asks user to enter Yes or No
// Processes and returns true if they answer Yes
// else returns false
// ---------------------------------------------------------------------
bool inquire(const char query[])
{
	char answer;

	do
	{
		cout << query << " [Yes or No]" << endl;
		cin >> answer;
		answer = toupper(answer);
		eatLine();
	}
	while ((answer != 'Y') && (answer != 'N'));
	
	return (answer == 'Y');
}

// ---------------------------------------------------------------------
// ---------------------------------------------------------------------
// end of file
